"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from contextlib import suppress
from pathlib import Path

import bpy

with suppress(ModuleNotFoundError):
    import AppKit
    import objc

from PySide2.QtGui import QIcon
from PySide2.QtCore import QObject

from .blender_application import BlenderApplication


class DarwinBlenderApplication(BlenderApplication):
    """
    Darwin (MACOS) Implementation of BlenderApplication
    """

    def __init__(self, *args, **kwargs):
        # OSX Specific - Needs to initialize first
        self._ns_window = self.__get_application_window() or None

        super().__init__(*args, **kwargs)

    def _get_application_hwnd(self) -> int:
        """
        This finds the blender application window and collects the
        handler window ID

        Returns int: Handler Window ID
        """

        # Check to ensure ns_window is set
        if self._ns_window is None:
            self._ns_window = self.__get_application_window()

        return objc.pyobjc_id(self._ns_window.contentView())

    @staticmethod
    def _get_application_icon() -> QIcon:
        """
        This finds the running blender process, extracts the blender icon from the blender.exe file on disk and saves it to the user's temp folder.
        It then creates a QIcon with that data and returns it.

        Returns QIcon: Application Icon
        """

        blender_path = bpy.app.binary_path
        contents_path = Path(blender_path).resolve().parent.parent
        icon_path = contents_path / "Resources" / "blender icon.icns"
        return QIcon(str(icon_path))

    @staticmethod
    def _get_application_window() -> AppKit.NSApp.mainWindow:
        """
        Specific to OSX; Main application window

        Returns: Main NSWindow of the application
        """

        ns_window = AppKit.NSApp.mainWindow()
        ns_window.setSharingType_(AppKit.NSWindowSharingReadWrite)
        return ns_window

    def _on_focus_object_changed(self, focus_object: QObject):
        """
        Args:
            focus_object: Object to track focus event
        """

        if focus_object is self.blender_widget:
            self._ns_window.makeKey()
            with bpy.context.temp_override(window=bpy.context.window_manager.windows[0]):
                bpy.ops.bqt.return_focus("INVOKE_DEFAULT")



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
import bqt
import bqt.focus
import os
import sys
import bpy
import PySide2.QtCore as QtCore
from PySide2.QtWidgets import QApplication
from PySide2.QtCore import QDir
import logging
from pathlib import Path


bl_info = {
        "name": "PySide2 Qt wrapper (bqt)",
        "description": "Enable PySide2 QtWidgets in Blender",
        "author": "tech-artists.org",
        "version": (1, 0),
        "blender": (2, 80, 0),
        # "location": "",
        # "warning": "", # used for warning icon and text in add-ons panel
        # "wiki_url": "http://my.wiki.url",
        # "tracker_url": "http://my.bugtracker.url",
        "support": "COMMUNITY",
        "category": "UI"
        }


# CORE FUNCTIONS #

def instantiate_application() -> "bqt.blender_applications.BlenderApplication":
    """
    Create an instance of Blender Application

    Returns BlenderApplication: Application Instance

    """
    # enable dpi scale, run before creating QApplication
    QApplication.setHighDpiScaleFactorRoundingPolicy(QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)
    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)
    image_directory = str(Path(__file__).parent / "images")
    QDir.addSearchPath('images', image_directory)
    app = QApplication.instance()
    if not app:
        app = load_os_module()
    return app


def load_os_module() -> "bqt.blender_applications.BlenderApplication":
    """Loads the correct OS platform Application Class"""
    operating_system = sys.platform
    if operating_system == "darwin":
        from .blender_applications.darwin_blender_application import DarwinBlenderApplication

        return DarwinBlenderApplication(sys.argv)

    if operating_system in ["linux", "linux2"]:
        # TODO: LINUX module
        pass

    elif operating_system == "win32":
        from .blender_applications.win32_blender_application import Win32BlenderApplication

        return Win32BlenderApplication(sys.argv)


@bpy.app.handlers.persistent
def create_global_app():
    """
    Create a global QApplication instance, that's maintained between Blender sessions.
    Runs after Blender finished startup.
    """
    # global qapp
    qapp = instantiate_application()

    # save a reference to the C++ window in a global var, to prevent the parent being garbage collected
    # for some reason this works here, but not in the blender_applications init as a class attribute (self),
    # and saving it in a global in blender_applications.py causes blender to crash on startup
    global parent_window
    parent_window = qapp._blender_window.parent()


def register():
    """
    Runs on enabling the add-on.
    setup bqt, wrap blender in qt, register operators
    """

    # hacky way to check if we already are waiting on bqt setup, or bqt is already setup
    if QApplication.instance():
        logging.warning("bqt: QApplication already exists, skipping bqt registration")
        return

    if os.getenv("BQT_DISABLE_STARTUP", 0) == "1":
        logging.warning("bqt: BQT_DISABLE_STARTUP is set, skipping bqt registration")
        return

    # only start focus operator if blender is wrapped
    if not os.getenv("BQT_DISABLE_WRAP", 0) == "1":
        # todo check if operator is already registered
        bpy.utils.register_class(bqt.focus.QFocusOperator)
        # append add_focus_handle before create_global_app, else it doesn't run on blender startup

    create_global_app()


def unregister():
    """
    Runs on disabling the add-on.
    """
    # todo, as long as blender is wrapped in qt, unregistering operator & callback will cause issues,
    #  for now we just return since unregister should not be called partially
    return

    # if not os.getenv("BQT_DISABLE_WRAP", 0) == "1":
    #     bpy.utils.unregister_class(bqt.focus.QFocusOperator)





"""
when alt tabbing in and out of blender, a bug happens when blender is wrapped in qt.
some keys stay stuck, e.g. alt or control, resulting in the user
not being able to use blender on refocus

this module fixes the bug by sending a key release event to the window
"""

import ctypes
import bpy


class QFocusOperator(bpy.types.Operator):
    bl_idname = "bqt.return_focus"  # access from bpy.ops.bqt.return_focus
    bl_label = "Fix bug related to bqt focus"
    bl_description = "Fix bug related to bqt focus"
    bl_options = {"INTERNAL"}

    def __init__(self):
        super().__init__()

    def __del__(self):
        """called when the operator finishes"""
        pass

    def invoke(self, context, event):
        """
        simulate key release to prevent blender from ignoring mouse input, after refocusing blender
        """
        self._detect_keyboard()
        return {"FINISHED"}

    @staticmethod
    def _detect_keyboard():
        """
        force a release of 'stuck' keys
        """

        # key codes from https://itecnote.com/tecnote/python-simulate-keydown/
        keycodes = [
            ("_ALT", 0x12),
            ("_CTRL", 0x11),
            ("_SHIFT", 0x10),
            ("VK_LWIN", 0x5B),
            ("VK_RWIN", 0x5C),
            ("OSKEY", 0x5B),  # dupe oskey, blender names it this
        ]

        # print("event.type", event.type, type(event.type))
        for name, code in keycodes:

            # todo this bug fix is not perfect yet, blender works better without this atm
            # # if the first key pressed is one of the following,
            # # don't simulate a key release, since it causes this bug:
            # # the first keypress on re-focus blender will be ignored, e.g. ctrl + v will just be v
            # if name in event.type:
            #     print("skipping:", name)
            #     continue

            # safely release all other keys that might be stuck down
            ctypes.windll.user32.keybd_event(code, 0, 2, 0)  # release key
            # print("released key", name, code)

        # todo, fix: blender occasionally still frozen input, despite having run the above code
        # but when we click the mouse, it starts working again
        # simulate a right mouse click did not work ...
        # ctypes.windll.user32.mouse_event(0x0008, 0, 0, 0, 0)  # right mouse click



"""
A demonstration script showing how to parent a QDialog to the Blender main window.
Yes, this could be done with QtWidgets.QMessageBox more easily.
The point is to show how to parent a window to the Blender application.

from bqt_demo import hello_world
hello_world.demo()
"""

import PySide2

Qt = PySide2.QtCore.Qt
QApplication = PySide2.QtWidgets.QApplication
QDialog = PySide2.QtWidgets.QDialog
QHBoxLayout = PySide2.QtWidgets.QHBoxLayout
QLabel = PySide2.QtWidgets.QLabel


class HelloWorldDialog(QDialog):
    """
    A sample 'Hello World!' QDialog.
    """

    def __init__(self, parent):
        super().__init__(parent, Qt.WindowCloseButtonHint | Qt.WindowSystemMenuHint | Qt.WindowTitleHint)

        self.resize(200, 50)
        self.setWindowTitle("Qt for Python in Blender")

        lbl_hw = QLabel("Hello World!")

        main_layout = QHBoxLayout()
        main_layout.addWidget(lbl_hw)
        self.setLayout(main_layout)


def demo():
    main_window = QApplication.instance().blender_widget
    dlg = HelloWorldDialog(main_window)
    dlg.show()


if __name__ == "__main__":
    demo()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
import sys
from .blender_application import BlenderApplication

if sys.platform == "darwin":
    from .darwin_blender_application import DarwinBlenderApplication
if sys.platform in ["linux", "linux2"]:
    # TODO: LINUX module
    pass
elif sys.platform == "win32":
    from .win32_blender_application import Win32BlenderApplication



"""
from bqt_demo import anim_bar
anim_bar.main()
"""
from PySide2 import QtWidgets
from PySide2.QtCore import Qt, QTimer
import bpy


class Window(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.Tool | Qt.WindowStaysOnTopHint)

        self.label1 = QtWidgets.QLabel("Changing the current frame in blender updates this slider.")
        self.label2 = QtWidgets.QLabel("Changing this slider updates the current frame in blender.")
        self.slider = QtWidgets.QSlider(Qt.Horizontal)

        with bpy.context.temp_override(window=bpy.context.window_manager.windows[0]):
            self.slider.setMinimum(bpy.context.scene.frame_start)
            self.slider.setMaximum(bpy.context.scene.frame_end)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.label1)
        self.layout().addWidget(self.label2)
        self.layout().addWidget(self.slider)

        self.timer = QTimer()
        self.timer.timeout.connect(self.on_update)

        self.slider.valueChanged.connect(self.slider_changed)

    def show(self):
        super(Window, self).show()
        tick = int(1000 / 30)  # tick 1000 / frames per second
        self.timer.start(tick)

    def on_update(self):
        """set slider to current frame from blender"""
        index = bpy.context.scene.frame_current
        self.slider.setValue(index)

    def slider_changed(self, value):
        # note that if this fails blender will crash, to prevent you can use try/except
        bpy.context.scene.frame_set(value)

        # TODO add click support, currently it only works when you drag the slider


def main():
    main_window = QtWidgets.QApplication.instance().blender_widget
    w = Window(main_window)
    w.show()
    return w



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from abc import abstractmethod, abstractstaticmethod, ABCMeta
from pathlib import Path
import os
from PySide2.QtWidgets import QApplication, QWidget
from PySide2.QtWidgets import QMainWindow
from PySide2.QtGui import QCloseEvent, QIcon, QImage, QPixmap, QWindow
from PySide2.QtCore import QEvent, QObject, QRect, QSettings
from bqt.ui.quit_dialogue import BlenderClosingDialog
import bpy


STYLESHEET_PATH = Path(__file__).parents[1] / "blender_stylesheet.qss"
ORGANISATION = "Tech-Artists.org"
APP = "Blender Qt"
WINDOW_TITLE = "Blender Qt"
WINDOW_GROUP_NAME = "MainWindow"
GEOMETRY = "Geometry"
MAXIMIZED = "IsMaximized"
FULL_SCREEN = "IsFullScreen"


class BlenderApplication(QApplication):
    """
    Base Implementation for QT Blender Window Container
    """

    def __init__(self, *args, **kwargs):
        __metaclass__ = ABCMeta
        super().__init__(*args, **kwargs)

        if STYLESHEET_PATH.exists():
            self.setStyleSheet(STYLESHEET_PATH.read_text())

        QApplication.setWindowIcon(self._get_application_icon())

        # Blender Window
        self._hwnd = self._get_application_hwnd()
        failed_to_get_handle = self._hwnd is None

        if os.getenv("BQT_DISABLE_WRAP") == "1" or failed_to_get_handle:
            self._blender_window = QWindow()
            self.blender_widget = QWidget.createWindowContainer(self._blender_window)
        else:
            self.blender_widget = QMainWindow()
            self.blender_widget.setWindowTitle(WINDOW_TITLE)
            self._blender_window = QWindow.fromWinId(self._hwnd)  # also sets flag to Qt.ForeignWindow
            self.window_container = QMainWindow.createWindowContainer(self._blender_window)
            self.blender_widget.setCentralWidget(self.window_container)
            self._set_window_geometry()
            self.blender_widget.show()
            self.focusObjectChanged.connect(self._on_focus_object_changed)

    @abstractstaticmethod
    def _get_application_hwnd() -> int:
        """
        This finds the blender application window and collects the
        handler window ID

        Returns int: Handler Window ID
        """

        return -1

    @staticmethod
    def _get_application_icon() -> QIcon:
        """
        This finds the running blender process, extracts the blender icon from the blender.exe file on disk and saves it to the user's temp folder.
        It then creates a QIcon with that data and returns it.

        Returns QIcon: Application Icon
        """

        icon_filepath = Path(__file__).parents[1] / "images" / "blender_icon_16.png"
        icon = QIcon()

        if icon_filepath.exists():
            image = QImage(str(icon_filepath))
            if not image.isNull():
                icon = QIcon(QPixmap().fromImage(image))

        return icon

    @abstractmethod
    def _on_focus_object_changed(self, focus_object: QObject):
        """
        Args:
            focus_object: Object to track focus event
        """

        pass

    def _unwrapped_window_geometry(self) -> QRect:
        """
        Get the window geometry from the Blender window before it was wrapped in a QWidgetContainer
        Run this before wrapping the window in a QWidgetContainer
        Returns QRect(x, y, width, height)
        """
        window = bpy.context.window_manager.windows[0]
        height, widht = window.height, window.width
        x = window.x
        y = window.y  # blender y relative from bottom of screen to bottom of Blender window
        # convert y to be relative from the top
        current_screen_rect = self.primaryScreen().availableGeometry()
        y = current_screen_rect.height() - y - height
        y += 56  # title bar offset
        return QRect(x, y, widht, height)


    def _set_window_geometry(self):
        """
        Loads stored window geometry preferences and applies them to the QWindow.
        .setGeometry() sets the size of the window minus the window frame.
        For this reason it should be set on self.blender_widget.
        """
        settings = QSettings(ORGANISATION, APP)
        settings.beginGroup(WINDOW_GROUP_NAME)
        fullscreen = settings.value(FULL_SCREEN, defaultValue=False, type=bool)
        maximized = settings.value(MAXIMIZED, defaultValue=False, type=bool)
        saved_geometry = settings.value(GEOMETRY)
        settings.endGroup()

        if fullscreen:
            self.blender_widget.showFullScreen()
            return

        if maximized:
            self.blender_widget.showMaximized()
            return


        unwrapped_geometry = self._unwrapped_window_geometry()  # maintain unwrapped window size & pos
        geometry = saved_geometry or unwrapped_geometry  # if no saved geometry, use previous blender window size
        self.blender_widget.setGeometry(geometry)  # setGeometry is relative to its parent


    def notify(self, receiver: QObject, event: QEvent) -> bool:
        """
        Args:
            receiver: Object to receive event
            event: Event

        Returns: bool
        """

        if isinstance(event, QCloseEvent) and receiver in (self.blender_widget, self._blender_window):
            # catch the close event when clicking close on the qt window,
            # ignore the event, and ask user if they want to close blender if unsaved changes.
            event.ignore()

            self.store_window_geometry()  # save qt window geometry, to restore on next launch

            if os.getenv("BQT_DISABLE_CLOSE_DIALOGUE") == "1":
                # this triggers the default blender close event, showing the save dialog if needed
                bpy.ops.wm.quit_blender({"window": bpy.context.window_manager.windows[0]}, "INVOKE_DEFAULT")
            else:
                closing_dialog = BlenderClosingDialog(self.blender_widget)
                closing_dialog.execute()

            return False

        return super().notify(receiver, event)

    def store_window_geometry(self):
        """
        Stores the current window geometry for the QWindow
        The .geometry() method on QWindow includes the size of the application minus the window frame.
        For that reason the _blender_widget should be used.
        """

        settings = QSettings(ORGANISATION, APP)
        settings.beginGroup(WINDOW_GROUP_NAME)
        settings.setValue(GEOMETRY, self.blender_widget.geometry())
        settings.setValue(MAXIMIZED, self.blender_widget.isMaximized())
        settings.setValue(FULL_SCREEN, self.blender_widget.isFullScreen())
        settings.endGroup()



"""
from bqt_demo import timer
timer.main()
"""
# sample from https://pythonpyqt.com/qtimer/

import sys
from PySide2.QtWidgets import QWidget, QPushButton, QApplication, QListWidget, QGridLayout, QLabel
from PySide2.QtCore import Qt, QTimer, QDateTime


class WinForm(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.Tool | Qt.WindowStaysOnTopHint)
        self.setWindowTitle("QTimer example")

        # self.listFile=QListWidget()
        self.label = QLabel("Label")
        self.startBtn = QPushButton("Start")
        self.endBtn = QPushButton("Stop")

        layout = QGridLayout()

        self.timer = QTimer()
        self.timer.timeout.connect(self.showTime)

        layout.addWidget(self.label, 0, 0, 1, 2)
        layout.addWidget(self.startBtn, 1, 0)
        layout.addWidget(self.endBtn, 1, 1)

        self.startBtn.clicked.connect(self.startTimer)
        self.endBtn.clicked.connect(self.endTimer)

        self.setLayout(layout)

    def showTime(self):
        time = QDateTime.currentDateTime()
        timeDisplay = time.toString("yyyy-MM-dd hh:mm:ss dddd")
        self.label.setText(timeDisplay)
        print(timeDisplay)

    def startTimer(self):
        self.timer.start(1000)
        self.startBtn.setEnabled(False)
        self.endBtn.setEnabled(True)

    def endTimer(self):
        self.timer.stop()
        self.startBtn.setEnabled(True)
        self.endBtn.setEnabled(False)


def main():
    # app=QApplication(sys.argv)
    main_window = QApplication.instance().blender_widget
    form = WinForm(main_window)
    form.show()
    # sys.exit(app.exec_())
    return form


if __name__ == "__main__":
    main()



import bpy
from PySide2.QtWidgets import QMessageBox
from PySide2.QtCore import Qt
import os
import bqt.ui


def shutdown_blender(*args):
    # By default changes to preferences are saved on exit, this can be toggled off in the preferences
    if bpy.context.preferences.use_preferences_save:
        bpy.ops.wm.save_userpref()

    bpy.ops.wm.quit_blender()


class WINDOW_OT_SaveFileFromQt(bpy.types.Operator):
    bl_idname = "wm.save_from_qt"
    bl_label = "Save_from_Qt"

    def execute(self, context):
        # TODO not sure what we are doing here, Friederman?
        if context.blend_data.is_saved:
            bpy.ops.wm.save_mainfile({"window": bpy.context.window_manager.windows[0]}, 'EXEC_AREA', check_existing=False)
        else:
            bpy.ops.wm.save_mainfile({"window": bpy.context.window_manager.windows[0]}, 'INVOKE_AREA', check_existing=False)
        # https://docs.blender.org/api/current/bpy.ops.html
        # EXEC_AREA - execute the operator in a certain context
        return {'FINISHED'}

# todo
#  when clicking the icon, the dialogue resets to center screen position
#  support dragging the dialogue around
#  add qshortcuts https://stackoverflow.com/questions/19845774/is-it-possible-to-use-an-underlined-letter-as-keyboard-shortcut-in-qt


class BlenderClosingDialog(QMessageBox):
    def __init__(self, parent):
        super().__init__(parent) #, Qt.WindowCloseButtonHint | Qt.WindowSystemMenuHint | Qt.WindowTitleHint | Qt.WindowStaysOnTopHint)

        # hide title bar
        self.setWindowFlag(Qt.FramelessWindowHint)

        filepath = bpy.data.filepath
        if not filepath:
            filepath = 'untitled.blend'
        filename = os.path.split(filepath)[1]

        question_icon = bqt.ui.get_question_pixmap()

        self.setText("Save changes before closing?\n\n" + filename)
        self.setStandardButtons(QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel)
        self.setDefaultButton(QMessageBox.Save)
        self.setIconPixmap(question_icon)

    def execute(self):
        if not bpy.data.is_dirty:
            shutdown_blender()
            return

        choice = super().exec_()
        if choice == QMessageBox.Save:
            bpy.utils.register_class(WINDOW_OT_SaveFileFromQt)
            bpy.app.handlers.save_post.append(shutdown_blender)
            bpy.ops.wm.save_from_qt()
        elif choice == QMessageBox.Discard:
            shutdown_blender()
        else:  # user clicked cancel
            pass
        return choice



"""
from bqt_demo import pyside_widgets
pyside_widgets.main()
"""

# https://www.pythonguis.com/tutorials/pyside-widgets/

from PySide2.QtCore import Qt
from PySide2.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDateEdit,
    QDateTimeEdit,
    QDial,
    QDoubleSpinBox,
    QFontComboBox,
    QLabel,
    QLCDNumber,
    QLineEdit,
    QMainWindow,
    QProgressBar,
    QPushButton,
    QRadioButton,
    QSlider,
    QSpinBox,
    QTimeEdit,
    QVBoxLayout,
    QWidget,
)


class Window(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.Tool | Qt.WindowStaysOnTopHint)

        self.setWindowTitle("Widgets App")

        layout = QVBoxLayout()
        widgets = [
            QCheckBox,
            QComboBox,
            QDateEdit,
            QDateTimeEdit,
            QDial,
            QDoubleSpinBox,
            QFontComboBox,
            QLCDNumber,
            QLabel,
            QLineEdit,
            QProgressBar,
            QPushButton,
            QRadioButton,
            QSlider,
            QSpinBox,
            QTimeEdit,
        ]

        for w in widgets:
            layout.addWidget(QLabel(w.__name__))
            layout.addWidget(w())

        widget = QWidget()
        widget.setLayout(layout)

        # Set the central widget of the Window. Widget will expand
        # to take up all the space in the window by default.
        self.setCentralWidget(widget)


def main():
    main_window = QApplication.instance().blender_widget
    w = Window(main_window)
    w.show()
    return w



"""
demo to show widget always on top
"""

import sys
from PySide2 import QtCore, QtWidgets


class mymainwindow(QtWidgets.QWidget):
    def __init__(self):
        super().__init__(parent=None, f=QtCore.Qt.WindowStaysOnTopHint)


app = QtWidgets.QApplication(sys.argv)
mywindow = mymainwindow()
mywindow.show()
app.exec_()



import bpy
import traceback


def try_except(func):
    """
    Prevent blender from crashing on an exception.
    Decorator to wrap a function in try except and print the traceback
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            traceback.print_exc()

    return wrapper


def context_window(func):
    """
    Support running operators from QT (ex. on button click).
    Decorator to override the context window for a function,
    """

    def wrapper(*args, **kwargs):
        with bpy.context.temp_override(window=bpy.context.window_manager.windows[0]):
            return func(*args, **kwargs)

    return wrapper
